import threading


class GVK(object):

    def __init__(self, group="", version="", kind="", response_type=""):
        self.group = group
        self.version = version
        self.kind = kind
        self.response_type = response_type

    @property
    def group_version(self):
        if self.group != "":
            return "%s/%s" % (self.group, self.version)
        return self.version


def get_resource_client(kube_client, gvk):
    return kube_client.resources.get(api_version=gvk.group_version, kind=gvk.kind)


ModelPackages = [
    "kubernetes.client.models",
    "dragonk8s.dragon.models"
]

ModelPackageModules = []

lock = threading.Lock()


def register_model_package(*package):
    with lock:
        for pkg in package:
            if pkg not in ModelPackages:
                ModelPackages.append(pkg)
        ModelPackageModules.clear()
        for pkg in ModelPackages:
            ModelPackageModules.append(__import__(pkg, fromlist=[""]))


AppsV1ControllerRevision = GVK(
    group="apps", version="v1", kind="ControllerRevision", response_type="V1ControllerRevision")

IoDragonAppsV1ReplicaSet = GVK(
    group="apps.dragon.io", version="v1", kind="ReplicaSet", response_type="IoDragonAppsV1ReplicaSet")

CoreV1Pod = GVK(
    group="", version="v1", kind="Pod", response_type="V1Pod")

EventsV1Event = GVK(
    group="events.k8s.io", version="v1", kind="Event", response_type="EventsV1Event")

ALL = [AppsV1ControllerRevision, IoDragonAppsV1ReplicaSet, CoreV1Pod, EventsV1Event]


def register_gvk(gvk: GVK):
    with lock:
        if gvk not in ALL:
            ALL.append(gvk)



import logging
import threading

from dragonk8s.lib.client.tools import informer, cache
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.client import CommonWatch
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager


def new_filtered_informer(
        client, gvk: apigvk.GVK, namespace: str, resync_period:int, indexers: dict, tweak_list_options, stop) -> informer.SharedIndexInformer:
    client = apigvk.get_resource_client(client, gvk)

    def _list_func(options: meta_v1.ListOptions) -> list:
        if tweak_list_options:
            tweak_list_options(options)
        return client.list(namespace=namespace, **options.to_params())

    def _watch_func(options: meta_v1.ListOptions):
        if tweak_list_options:
            tweak_list_options(options)
        return client.watch(namespace=namespace, watcher=CommonWatch(client), **options.to_params())
    return informer._SharedIndexInformer(
        lw=informer.ListWatch(list_func=_list_func, watch_func=_watch_func),
        example_obj=gvk.response_type,
        default_event_handler_resync_period=resync_period,
        indexers=indexers,
        stop=stop
    )


def new_informer(client, gvk: apigvk.GVK, namespace: str, resync_period:int, indexers: dict, stop) -> informer.SharedIndexInformer:
    return new_filtered_informer(client, gvk, namespace, resync_period, indexers, tweak_list_options=None, stop=stop)


class SharedInformerFactory(object):

    def __init__(self, client, default_resync, namespace="", tweak_list_options=None, stop=None):
        self.client = client
        self.default_resync = default_resync
        self.namespace = namespace
        self.tweak_list_options = tweak_list_options
        self.informers = {}
        self.started_informers = {}
        self._custom_resync = {}
        self._lock = threading.Lock()
        self.stop = stop

    def start(self, stop: threading.Event):
        with self._lock:
            for informer_type, ifm in self.informers.items():
                if informer_type not in self.started_informers:
                    logging.info("start informer for %s" % informer_type)
                    th = GlobalThreadManager.new_thread(target=ifm.run, args=(stop, ), name="informer-%s" % informer_type)
                    th.start()
                    self.started_informers[informer_type] = th

    def informer_for(self, gvk: apigvk.GVK, **kwargs) -> informer.SharedIndexInformer:
        with self._lock:
            informer_type = gvk.response_type
            if informer_type in self.informers:
                return self.informers[informer_type]
        resync_period = self.default_resync
        if informer_type in self._custom_resync:
            resync_period = self._custom_resync[informer_type]
        if "indexers" in kwargs:
            indexers = kwargs["indexers"]
        else:
            indexers = {}
        namespace = self.namespace
        if "namespace" in kwargs:
            namespace = kwargs["namespace"]
        tweak_list_options = None
        if "tweak_list_options" in kwargs:
            tweak_list_options = kwargs["tweak_list_options"]
        ifm = new_filtered_informer(client=self.client,
                                    gvk=gvk,
                                    resync_period=resync_period,
                                    indexers=indexers,
                                    tweak_list_options=tweak_list_options,
                                    namespace=namespace,
                                    stop=self.stop)

        self.informers[informer_type] = ifm
        return ifm

    def wait_for_cache_sync(self, stop: threading.Event) -> dict:
        with self._lock:
            informers = {}
            for informer_type, ifm in self.informers.items():
                if informer_type in self.started_informers:
                    informers[informer_type] = ifm
            res = {}
            for informer_type, ifm in informers.items():
                res[informer_type] = cache.wait_for_cache_sync(stop, ifm.has_synced)
            return res

    def set_custom_resync(self, resync_config: dict):
        self._custom_resync = resync_config



import threading
import time
import logging
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager
from threading import Event


def until(f, period: int, stop_event: Event):
    count = 0
    while not stop_event.is_set():
        if count == period:
            f()
            count = 0
        else:
            count += 1
            time.sleep(1)


def until_named(f, period: int, stop_event: Event, name):
    count = 0
    while not stop_event.is_set():
        if count == period:
            f()
            count = 0
        else:
            count += 1
            time.sleep(1)
    logging.info("stop until %s" % name)


def until_with_thread(f, period: int, stop_event: Event, name):
    def do(df, dperiod, dstop_event):
        until_named(df, dperiod, dstop_event, name)
    backend = GlobalThreadManager.new_thread(generate_name="until_with_thread_%s" % name, target=do, args=(f, period, stop_event))
    backend.start()


class TimeOutException(Exception):
    pass


def poll(f, period: int, timeout: int, stop: threading.Event, immadiate=False) -> bool:
    if timeout < 1:
        timeout = 0
    t = 0
    ok = False
    while not stop.is_set():
        if immadiate:
            ok = f()
        if t % period == 0:
            ok = f()
        if ok:
            return True
        time.sleep(1)
        t += 1
        if 0 < timeout < t:
            return False
    logging.info("stop poll %s" % f.__name__)
    return False





class Operator(object):
    DoesNotExist = "!"
    Equals = "="
    DoubleEquals = "=="
    In = "in"
    NotEquals = "!="
    NotIn = "notin"
    Exists = "exists"
    GreaterThan = "gt"
    LessThan = "lt"



# coding: utf-8

"""
    Kubernetes

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: v1.24.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from dragonk8s.dragon.configuration import Configuration


class IoDragonAppsV1JobSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'active_deadline_seconds': 'int',
        'backoff_limit': 'int',
        'completion_mode': 'str',
        'completions': 'int',
        'manual_selector': 'bool',
        'parallelism': 'int',
        'selector': 'IoDragonAppsV1JobSpecSelector',
        'suspend': 'bool',
        'template': 'IoDragonAppsV1JobSpecTemplate',
        'ttl_seconds_after_finished': 'int'
    }

    attribute_map = {
        'active_deadline_seconds': 'activeDeadlineSeconds',
        'backoff_limit': 'backoffLimit',
        'completion_mode': 'completionMode',
        'completions': 'completions',
        'manual_selector': 'manualSelector',
        'parallelism': 'parallelism',
        'selector': 'selector',
        'suspend': 'suspend',
        'template': 'template',
        'ttl_seconds_after_finished': 'ttlSecondsAfterFinished'
    }

    def __init__(self, active_deadline_seconds=None, backoff_limit=None, completion_mode=None, completions=None, manual_selector=None, parallelism=None, selector=None, suspend=None, template=None, ttl_seconds_after_finished=None, local_vars_configuration=None):  # noqa: E501
        """IoDragonAppsV1JobSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._active_deadline_seconds = None
        self._backoff_limit = None
        self._completion_mode = None
        self._completions = None
        self._manual_selector = None
        self._parallelism = None
        self._selector = None
        self._suspend = None
        self._template = None
        self._ttl_seconds_after_finished = None
        self.discriminator = None

        if active_deadline_seconds is not None:
            self.active_deadline_seconds = active_deadline_seconds
        if backoff_limit is not None:
            self.backoff_limit = backoff_limit
        if completion_mode is not None:
            self.completion_mode = completion_mode
        if completions is not None:
            self.completions = completions
        if manual_selector is not None:
            self.manual_selector = manual_selector
        if parallelism is not None:
            self.parallelism = parallelism
        if selector is not None:
            self.selector = selector
        if suspend is not None:
            self.suspend = suspend
        self.template = template
        if ttl_seconds_after_finished is not None:
            self.ttl_seconds_after_finished = ttl_seconds_after_finished

    @property
    def active_deadline_seconds(self):
        """Gets the active_deadline_seconds of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The active_deadline_seconds of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: int
        """
        return self._active_deadline_seconds

    @active_deadline_seconds.setter
    def active_deadline_seconds(self, active_deadline_seconds):
        """Sets the active_deadline_seconds of this IoDragonAppsV1JobSpec.


        :param active_deadline_seconds: The active_deadline_seconds of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: int
        """

        self._active_deadline_seconds = active_deadline_seconds

    @property
    def backoff_limit(self):
        """Gets the backoff_limit of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The backoff_limit of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: int
        """
        return self._backoff_limit

    @backoff_limit.setter
    def backoff_limit(self, backoff_limit):
        """Sets the backoff_limit of this IoDragonAppsV1JobSpec.


        :param backoff_limit: The backoff_limit of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: int
        """

        self._backoff_limit = backoff_limit

    @property
    def completion_mode(self):
        """Gets the completion_mode of this IoDragonAppsV1JobSpec.  # noqa: E501

        CompletionMode specifies how Pod completions of a Job are tracked.  # noqa: E501

        :return: The completion_mode of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: str
        """
        return self._completion_mode

    @completion_mode.setter
    def completion_mode(self, completion_mode):
        """Sets the completion_mode of this IoDragonAppsV1JobSpec.

        CompletionMode specifies how Pod completions of a Job are tracked.  # noqa: E501

        :param completion_mode: The completion_mode of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: str
        """

        self._completion_mode = completion_mode

    @property
    def completions(self):
        """Gets the completions of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The completions of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: int
        """
        return self._completions

    @completions.setter
    def completions(self, completions):
        """Sets the completions of this IoDragonAppsV1JobSpec.


        :param completions: The completions of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: int
        """

        self._completions = completions

    @property
    def manual_selector(self):
        """Gets the manual_selector of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The manual_selector of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: bool
        """
        return self._manual_selector

    @manual_selector.setter
    def manual_selector(self, manual_selector):
        """Sets the manual_selector of this IoDragonAppsV1JobSpec.


        :param manual_selector: The manual_selector of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: bool
        """

        self._manual_selector = manual_selector

    @property
    def parallelism(self):
        """Gets the parallelism of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The parallelism of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: int
        """
        return self._parallelism

    @parallelism.setter
    def parallelism(self, parallelism):
        """Sets the parallelism of this IoDragonAppsV1JobSpec.


        :param parallelism: The parallelism of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: int
        """

        self._parallelism = parallelism

    @property
    def selector(self):
        """Gets the selector of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The selector of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: IoDragonAppsV1JobSpecSelector
        """
        return self._selector

    @selector.setter
    def selector(self, selector):
        """Sets the selector of this IoDragonAppsV1JobSpec.


        :param selector: The selector of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: IoDragonAppsV1JobSpecSelector
        """

        self._selector = selector

    @property
    def suspend(self):
        """Gets the suspend of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The suspend of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: bool
        """
        return self._suspend

    @suspend.setter
    def suspend(self, suspend):
        """Sets the suspend of this IoDragonAppsV1JobSpec.


        :param suspend: The suspend of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: bool
        """

        self._suspend = suspend

    @property
    def template(self):
        """Gets the template of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The template of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: IoDragonAppsV1JobSpecTemplate
        """
        return self._template

    @template.setter
    def template(self, template):
        """Sets the template of this IoDragonAppsV1JobSpec.


        :param template: The template of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: IoDragonAppsV1JobSpecTemplate
        """
        if self.local_vars_configuration.client_side_validation and template is None:  # noqa: E501
            raise ValueError("Invalid value for `template`, must not be `None`")  # noqa: E501

        self._template = template

    @property
    def ttl_seconds_after_finished(self):
        """Gets the ttl_seconds_after_finished of this IoDragonAppsV1JobSpec.  # noqa: E501


        :return: The ttl_seconds_after_finished of this IoDragonAppsV1JobSpec.  # noqa: E501
        :rtype: int
        """
        return self._ttl_seconds_after_finished

    @ttl_seconds_after_finished.setter
    def ttl_seconds_after_finished(self, ttl_seconds_after_finished):
        """Sets the ttl_seconds_after_finished of this IoDragonAppsV1JobSpec.


        :param ttl_seconds_after_finished: The ttl_seconds_after_finished of this IoDragonAppsV1JobSpec.  # noqa: E501
        :type: int
        """

        self._ttl_seconds_after_finished = ttl_seconds_after_finished

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, IoDragonAppsV1JobSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, IoDragonAppsV1JobSpec):
            return True

        return self.to_dict() != other.to_dict()



import string


class GroupKind(object):

    def __init__(self, group="", kind=""):
        self.group = group
        self.kind = kind

    def empty(self) -> bool:
        return self.group == "" and self.kind == ""

    def __str__(self):
        if self.group == "":
            return self.kind
        return "%s.%s" %(self.kind, self.group)


class GroupVersionKind(object):

    def __init__(self, group="", version="", kind=""):
        self.group = group
        self.version = version
        self.kind = kind

    def empty(self) -> bool:
        return self.group == "" and self.version == "" and self.kind == ""

    def __str__(self):
        return "%s/%s, kind=%s" % (self.group, self.version, self.kind)


class GroupVersion(object):

    def __init__(self, group="", version=""):
        self.group = group
        self.version = version

    def empty(self) -> bool:
        return self.group == "" and self.version == ""

    def __str__(self):
        if self.group != "":
            return "%s/%s" % (self.group, self.version)
        return self.version

    def identifier(self):
        return str(self)


def parse_group_version(gv: str) -> GroupVersion:
    if len(gv) == 0 or gv == "/":
        return GroupVersion()
    c = gv.count("/")
    if c == 0:
        return GroupVersion(group="", version=gv)
    if c == 1:
        i = gv.index("/")
        return GroupVersion(group=gv[:i], version=gv[i+1:])
    raise Exception("unexpected GroupVersion string: %s", gv)


def from_api_version_and_kind(api_version: str, kind: str) -> GroupVersionKind:
    try:
        gv = parse_group_version(api_version)
        return GroupVersionKind(group=gv.group, version=gv.version, kind=kind)
    except:
        return GroupVersionKind(kind=kind)


def from_obj(obj) -> GroupVersionKind:
    if obj is None:
        return GroupVersionKind()
    api_version = ""
    kind = ""
    if hasattr(obj, "api_version"):
        api_version = obj.api_version
    if hasattr(obj, "kind"):
        kind = obj.kind
    return from_api_version_and_kind(api_version, kind)



import copy
import threading

import time

import logging

from kubernetes.client.models.v1_pod_template_spec import V1PodTemplateSpec
from kubernetes.client.models.v1_owner_reference import V1OwnerReference
from dragonk8s.lib.client.tools import cache
from kubernetes import client
from dragonk8s.lib.client.tools import events
from kubernetes.client.models import V1ObjectMeta
from kubernetes.client.models import V1Pod
from kubernetes.client import models as kube_models
from kubernetes.client.exceptions import ApiException
from dragonk8s.lib.apimachinery.pkg.api import errors
from dragonk8s.lib.apimachinery.pkg.types.patch import *
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.models import corev1
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.apimachinery.pkg import labels


ExpectationsTimeout = 5 * 60
FailedCreatePodReason = "FailedCreate"
SuccessfulCreatePodReason = "SuccessfulCreate"
FailedDeletePodReason = "FailedDelete"
SuccessfulDeletePodReason = "SuccessfulDelete"


class PodControlInterface(object):

    def create_pod(self, namespace: str, template: V1PodTemplateSpec, obj, controller_ref: V1OwnerReference):
        pass

    def create_pod_with_generate_name(self, namespace: str, template: V1PodTemplateSpec, obj,
                                      controller_ref: V1OwnerReference, generate_name: str):
        pass

    def delete_pod(self, namespace: str, pod_id: str, obj):
        pass

    def patch_pod(self, namespace: str, name: str, data: dict):
        pass


class ControlleeExpectations(object):

    def __init__(self, add: int, dele: int, key: str, timestamp: float):
        self.addn = add
        self.deln = dele
        self.key = key
        self.timestamp = timestamp

    def add(self, addn: int, deln: int):
        self.addn += addn
        self.deln = deln

    def fulfilled(self) -> bool:
        return self.addn <= 0 and self.deln <= 0

    def get_expectations(self) -> (int, int):
        return self.addn, self.deln

    def is_expired(self):
        return time.time() - self.timestamp > ExpectationsTimeout

    def __repr__(self):
        return "(key=%s, add=%d, del=%d, timestamp=%d)" % (self.key, self.addn, self.deln, self.timestamp)

    def __str__(self):
        return self.__repr__()


class ControllerExpectationsInterface(object):

    def get_expectations(self, controllerkey: str) -> (ControlleeExpectations, bool):
        pass

    def satisfied_expectations(self, controllerkey: str) -> bool:
        pass

    def delete_expectations(self, controllerkey: str):
        pass

    def set_expectations(self, controllerkey: str, addn: int, deln: int):
        pass

    def expect_creations(self, controllerkey: str, addns: int):
        pass

    def expect_deletions(self, controllerkey: str, delns: int):
        pass

    def creation_observed(self, controllerkey: str):
        pass

    def deletion_observed(self, controllerkey: str):
        pass

    def raise_expectations(self, controllerkey: str, addn: int, deln: int):
        pass

    def lower_expectations(self, controllerkey: str, addn: int, deln: int):
        pass


def exp_key_func(obj) -> str:
    return obj.key


class ControllerExpectations(ControllerExpectationsInterface):
    def __init__(self):
        super().__init__()
        self.cache = cache.Cache(key_func=exp_key_func, indexers={})

    def get_expectations(self, controllerkey: str) -> (ControlleeExpectations, bool):
        exp, exists = self.cache.get_by_key(controllerkey)
        if exists:
            return exp, True
        return None, False

    def satisfied_expectations(self, controllerkey: str) -> bool:
        exp, exists = self.cache.get_by_key(controllerkey)
        if not exists:
            logging.info("Controller %s either never recorded expectations, or the ttl expired" % controllerkey)
        else:
            if exp.fulfilled():
                logging.info("Controller expectations fulfilled: %s" % exp)
                return True
            elif exp.is_expired():
                logging.info("Controller expectations expired %s" % exp)
                return True
            else:
                logging.info("Controller still waiting on expectations %s" % exp)
                return False
        return True

    def delete_expectations(self, controllerkey: str):
        exp, exists = self.cache.get_by_key(controllerkey)
        if exists:
            self.cache.delete(exp)

    def set_expectations(self, controllerkey: str, addn: int, deln: int):
        exp = ControlleeExpectations(add=addn, dele=deln, key=controllerkey, timestamp=time.time())
        logging.info("Setting expectations %s" % exp)
        return self.cache.add(exp)

    def expect_creations(self, controllerkey: str, addns: int):
        return self.set_expectations(controllerkey, addns, 0)

    def expect_deletions(self, controllerkey: str, delns: int):
        return self.set_expectations(controllerkey, 0, delns)

    def creation_observed(self, controllerkey: str):
        self.lower_expectations(controllerkey, 1, 0)

    def deletion_observed(self, controllerkey: str):
        self.lower_expectations(controllerkey, 0, 1)

    def raise_expectations(self, controllerkey: str, addn: int, deln: int):
        exp, exists = self.get_expectations(controllerkey)
        if exists:
            exp.add(addn, deln)
            logging.info("raise expectations %s" % exp)

    def lower_expectations(self, controllerkey: str, addn: int, deln: int):
        exp, exists = self.get_expectations(controllerkey)
        if exists:
            exp.add(-1 * addn, -1 * deln)
            logging.info("lowered expectations %s" % exp)


def _validate_controller_ref(controller_ref: V1OwnerReference):
    if controller_ref is None:
        raise Exception("controllerRef is None")
    if len(controller_ref.api_version) == 0:
        raise Exception("controllerRef has empty api version")
    if len(controller_ref.kind) == 0:
        raise Exception("controllerRef has empty kind")
    if not controller_ref.controller:
        raise Exception("controllerRef.Controller is not set to true")
    if not controller_ref.block_owner_deletion:
        raise Exception("controllerRef.BlockOwnerDeletin is not set")


def _get_pods_label_set(template: V1PodTemplateSpec) -> dict:
    desired_labels = dict()
    if not template.metadata.labels:
        return desired_labels
    for k, v in template.metadata.labels.items():
        desired_labels[k] = v
    return desired_labels


def _get_pods_finalizers(template: V1PodTemplateSpec) -> list:
    desired_finalizers = copy.copy(template.metadata.finalizers)
    return desired_finalizers


def _get_pods_annotation_set(template: V1PodTemplateSpec) -> dict:
    desired_annotations = dict()
    if not template.metadata.annotations:
        return desired_annotations
    for k, v in template.metadata.annotations.items():
        desired_annotations[k] = v
    return desired_annotations


def get_pod_from_template(template: V1PodTemplateSpec, parent_object, controller_ref: V1OwnerReference) -> V1Pod:
    desired_labels = _get_pods_label_set(template)
    desired_finalizers = _get_pods_finalizers(template)
    desired_annotations = _get_pods_annotation_set(template)
    prefix = _get_pods_prefix(parent_object.metadata.name)
    pod = V1Pod(
        metadata=V1ObjectMeta(
            labels=desired_labels,
            annotations=desired_annotations,
            generate_name=prefix,
            finalizers=desired_finalizers,
        )
    )
    if controller_ref is not None:
        pod.metadata.owner_references = [copy.deepcopy(controller_ref)]
    pod.spec = copy.deepcopy(template.spec)
    return pod


def _get_pods_prefix(controller_name: str) -> str:
    prefix = "%s-" % controller_name
    # todo: validate
    return prefix


class RealPodControl(PodControlInterface):

    def __init__(self, kube_client, recorder: events.EventRecorder):
        self.kube_client = apigvk.get_resource_client(kube_client, apigvk.CoreV1Pod)
        self.recorder = recorder

    def _create_pod(self, namespace: str, pod: V1Pod, obj):
        if not pod.metadata.labels:
            raise Exception("unable to create pods , no labels")
        pod.metadata.namespace = namespace
        try:
            new_pod = self.kube_client.create(pod)
        except ApiException as e:
            if errors.has_status_cause(e, errors.NamespaceTerminatingCause):
                self.recorder.eventf(obj, None, events.EventTypeWarning, FailedCreatePodReason,
                                     "CreatePod", "Error creating: %s", str(e))
            raise e
        logging.info("Controller %s created pod %s", obj.metadata.name, new_pod.metadata.name)
        self.recorder.eventf(obj, None, events.EventTypeNormal, SuccessfulCreatePodReason,
                             SuccessfulCreatePodReason, "created pod: %s" % new_pod.metadata.name)

    def create_pod(self, namespace: str, template: V1PodTemplateSpec, obj, controller_ref: V1OwnerReference):
        self.create_pod_with_generate_name(namespace, template, obj, controller_ref, "")

    def create_pod_with_generate_name(self, namespace: str, template: V1PodTemplateSpec, obj,
                                      controller_ref: V1OwnerReference, generate_name: str):
        _validate_controller_ref(controller_ref)
        pod = get_pod_from_template(template, obj, controller_ref)
        if generate_name:
            pod.metadata.generate_name = generate_name
        return self._create_pod(namespace, pod, obj)

    def delete_pod(self, namespace: str, pod_id: str, obj):
        logging.info("deleting pod(%s) by controller %s" % (pod_id, obj.metadata.name))
        try:
            self.kube_client.delete(name=pod_id, namespace=namespace)
        except ApiException as e:
            if errors.is_not_found(e):
                logging.info("pod %s/%s has already been deleted" % (namespace, pod_id))
                raise e
            self.recorder.eventf(obj, None, events.EventTypeWarning,
                                 FailedDeletePodReason, FailedDeletePodReason, "Error deleting: %s" % str(e))
            raise Exception("unable to delete pods: %s" % str(e))
        except Exception as e:
            self.recorder.eventf(obj, None, events.EventTypeWarning,
                                 FailedDeletePodReason, FailedDeletePodReason, "Error deleting: %s" % str(e))
            raise Exception("unable to delete pods: %s" % str(e))
        self.recorder.eventf(obj, None, events.EventTypeNormal,
                             SuccessfulDeletePodReason, SuccessfulDeletePodReason, "Deleted pod: %s" % pod_id)

    def patch_pod(self, namespace: str, name: str, data: dict):
        self.kube_client.patch(body=data, name=name, namespace=namespace, content_type=PatchType.StrategicMergePatchType)


class UIDSet(object):
    def __init__(self, data, key):
        self.data = data
        self.key = key


def uid_set_key_func(obj: UIDSet) -> str:
    return obj.key


class UIDTrackingControllerExpectations(object):

    def __init__(self, ce: ControllerExpectationsInterface):
        self.ce = ce
        self.lock = threading.Lock()
        self.uid_store = cache.Cache(key_func=uid_set_key_func, indexers={})

    def get_uids(self, controller_key: str) -> list:
        uid, exists = self.uid_store.get_by_key(controller_key)
        if exists:
            return uid.data
        return []

    def expect_deletions(self, rc_key, deleted_keys):
        expected_uids = list(set(deleted_keys))
        existing = self.get_uids(rc_key)
        if len(existing) > 0:
            logging.error("Clobbering existing delete keys: %s" % str(existing))
        self.uid_store.add(UIDSet(data=expected_uids, key=rc_key))
        self.ce.expect_deletions(rc_key, len(expected_uids))

    def deletion_observed(self, rc_key, delete_key):
        try:
            self.lock.acquire()
            uids = self.get_uids(rc_key)
            print("uids: %s" % str(uids))
            if delete_key in uids:
                logging.info("Controller %s received delete for pod %s" % (rc_key, delete_key))
                self.ce.deletion_observed(rc_key)
                uids.remove(delete_key)
        finally:
            self.lock.release()

    def delete_expectations(self, rc_key):
        try:
            self.lock.acquire()
            self.ce.delete_expectations(rc_key)
            uid_exp, exists = self.uid_store.get_by_key(rc_key)
            if exists:
                self.uid_store.delete(uid_exp)

        finally:
            self.lock.release()


class RSControlInterface(object):

    def patch_replica_set(self, namespace: str, name: str, data: dict):
        pass


class RealRSControl(RSControlInterface):

    def __init__(self, client, recorder):
        self.kube_client = client
        self.recorder = recorder

    def patch_replica_set(self, namespace: str, name: str, data: dict):
        self.kube_client.patch(body=data, name=name, namespace=namespace, content_type=PatchType.MergePatchType)


class ControllerRevisionControlInterface(object):

    def patch_controller_revision(self, namespace: str, name: str, data: dict):
        pass


class RealControllerRevisionControl(ControllerRevisionControlInterface):

    def __init__(self, kube_client):
        self.kube_client = kube_client

    def patch_controller_revision(self, namespace: str, name: str, data: dict):
        apigvk.get_resource_client(self.kube_client, apigvk.AppsV1ControllerRevision).patch(
            body=data, name=name, namespace=namespace, content_type=PatchType.StrategicMergePatchType)


def is_pod_active(p: V1Pod) -> bool:
    return p.status.phase not in (corev1.PodStatus.PodSucceeded, corev1.PodStatus.PodFailed) and \
           p.metadata.deletion_timestamp is None


def filter_active_pods(pods: list) -> list:
    res = []
    for p in pods:
        if is_pod_active(p):
            res.append(p)
        else:
            logging.info("Ignoring inactive pod %s/%s in state %s, deletion time: %s"
                         % (p.metadata.namespace, p.metadata.name, p.status.phase, p.metadata.deletion_timestamp))
    return res


def recheck_deletion_timestamp(get_object):
    def do():
        try:
            obj = get_object()
        except Exception as e:
            raise Exception("can't recheck deletion_timestamp: %s" % str(e))
        if obj.metadata.deletion_timestamp is not None:
            raise Exception("%s/%s has just been deleted at: %s"
                            % (obj.metadata.namespace, obj.metadata.name, obj.metadata.deletion_timestamp))
        return
    return do


class BaseControllerRefManager(object):

    def __init__(self, controller, selector, can_adopt_func, lock: threading.Lock):
        self.controller = controller
        self.selector = selector
        self.can_adopt_exception = None
        self.can_adopt_func = can_adopt_func
        self.lock = lock

    def can_adopt(self):
        with self.lock:
            if self.can_adopt_func:
                try:
                    self.can_adopt_func()
                except Exception as e:
                    self.can_adopt_exception = e
                    raise self.can_adopt_exception

    def claim_object(self, obj, match, adopt, release) -> bool:
        controller_ref = meta_v1.get_controller_of_no_copy(obj)
        if controller_ref:
            if controller_ref.uid != self.controller.metadata.uid:
                return False
            if match(obj):
                return True
            if self.controller.metadata.deletion_timestamp:
                return False
            try:
                release(obj)
            except errors.ApiException as e:
                if errors.is_not_found(e):
                    return False
            return False
        if self.controller.metadata.deletion_timestamp or not match(obj):
            return False
        if obj.metadata.deletion_timestamp:
            return False
        if len(self.controller.metadata.namespace) > 0 and self.controller.metadata.namespace != obj.metadata.namespace:
            return False
        try:
            adopt(obj)
        except errors.ApiException as e:
            if errors.is_not_found(e):
                return False
        return True


def _owner_ref_controller_patch(controller, controller_kind: apigvk.GVK, uid: str, *finalizers) -> dict:
    res = dict(
        metadata=dict(
            uid=uid,
            ownerReferences=[
                dict(
                    apiVersion=controller_kind.version,
                    kind=controller_kind.kind,
                    uid=controller.metadata.uid,
                    controller=True,
                    blockOwnerDeletion=True,
                )
            ],
            finalizers=finalizers
        )
    )
    return res


def _owner_reference(uid: str, patch_type) -> dict:
    return {
        "$patch": patch_type,
        "uid": uid,
    }


def generate_delete_owner_ref_strategic_merge_data(dependent_uid: str, owner_uids: list, *finalizers) -> dict:
    owner_references = []
    for owner_uid in owner_uids:
        owner_references.append(_owner_reference(owner_uid, "delete"))
    patch = {
        "metadata": {
            "uid": dependent_uid,
            "ownerReferences": owner_references,
            "$deleteFromPrimitiveList/finalizers": finalizers,
        }
    }
    return patch


class PodControllerRefManager(BaseControllerRefManager):

    def __init__(self, pod_control: PodControlInterface,
                 controller, selector, controller_kind: apigvk.GVK, can_adopt_func, *finalizers):
        super(PodControllerRefManager, self).__init__(controller, selector, threading.Lock(), can_adopt_func)
        self.controller_kind = controller_kind
        self.pod_control = pod_control
        self.finalizers = finalizers

    def adopt_pod(self, pod: V1Pod):
        try:
            self.can_adopt()
        except Exception as e:
            raise Exception("can't adopt pod %s/%s (%s): %s"
                            % (pod.metadata.namespace, pod.metadata.name, pod.metadata.uid, str(e)))
        patch_data = _owner_ref_controller_patch(
            self.controller, self.controller_kind, pod.metadata.uid, *self.finalizers)
        self.pod_control.patch_pod(pod.metadata.namespace, pod.metadata.name, patch_data)

    def release_pod(self, pod: V1Pod):
        logging.info("patching pod %s_%s to remove its controllerRef to %s/%s:%s"
                     % (pod.metadata.namespace, pod.metadata.name, self.controller_kind.group_version,
                        self.controller_kind.kind, self.controller.metadata.name))
        patch_data = generate_delete_owner_ref_strategic_merge_data(
            pod.metadata.uid, [self.controller.metadata.uid], *self.finalizers)
        try:
            self.pod_control.patch_pod(pod.metadata.namespace, pod.metadata.name, patch_data)
        except errors.ApiException as e:
            if errors.is_not_found(e):
                return
            if errors.is_invalid(e):
                return

    def claim_pods(self, pods: list, *filters) -> list:
        claimed = []

        def match(obj) -> bool:
            pod = obj
            if not self.selector.matches(labels.Set(pod.metadata.labels)):
                return False
            for filter in filters:
                if not filter(pod):
                    return False
            return True

        for pod in pods:
            if self.claim_object(pod, match, self.adopt_pod, self.release_pod):
                claimed.append(pod)
        return claimed


def pod_key(pod: V1Pod) -> str:
    return "{}/{}".format(pod.metadata.namespace, pod.metadata.name)



from dragonk8s.lib.client.informers import factory
from dragonk8s.lib.client.tools import informer, cache
from dragonk8s.lib.client import lister
from dragonk8s.lib.apimeta import apigvk


class BaseInformer(object):

    def __init__(self, gvk: apigvk.GVK, informer_factory: factory.SharedInformerFactory, tweak_list_options=None):
        self.factory = informer_factory
        self.tweak_list_options = tweak_list_options
        self.gvk = gvk

    def informer(self) -> informer.SharedIndexInformer:
        return self.factory.informer_for(self.gvk,
                                         indexers={cache.NamespaceIndex: cache.meta_namespace_index_func},
                                         tweak_list_options=self.tweak_list_options)

    def lister(self):
        return lister.Lister(indexer=self.informer().get_indexer())


def get_informer(
        gvk: apigvk.GVK, informer_factory: factory.SharedInformerFactory, tweak_list_options=None) -> BaseInformer:
    return BaseInformer(gvk, informer_factory, tweak_list_options)



from distutils.core import setup
from setuptools import find_packages

setup(
    name='dragonk8s',
    version='1.0.0',
    description="A k8s implement of Python3",
    author='ma',
    author_email='591867837@qq.com',
    url='https://dragonk8s.com',
    packages=find_packages(),
)